"""Scenario manifest plumbing (adapted from Survi, sans heavy dependencies)."""
from __future__ import annotations

from dataclasses import asdict, dataclass
from enum import Enum
from pathlib import Path
from typing import Any, Iterable, Literal, Mapping

import numpy as np
import pandas as pd
from shapely.geometry import Polygon, Point

from .geom import SamplingParams, SurveyArea, make_grid, nearest_surface_values
from .rng import numpy_rng
from .synthetic import BumpSpec, RippleSpec, SurfaceParams, make_surface_df, make_rectangle, make_l_shape, rotate_polygon
from .units import Unit


class PolygonKind(str, Enum):
    RECT = "rect"
    ROTATED_RECT = "rotated_rect"
    L_SHAPE = "lshape"
    RANDOM_CONVEX = "random_convex"


@dataclass(frozen=True)
class GridSpec:
    grid_spacing: float = 3.0
    edge_exclusion: float = 1.5
    rotation_deg: float = 0.0
    shift_x: float = 0.0
    shift_y: float = 0.0

    def to_params(self) -> SamplingParams:
        return SamplingParams(
            shift_x=self.shift_x,
            shift_y=self.shift_y,
            rotation_deg=self.rotation_deg,
            grid_spacing=self.grid_spacing,
            edge_exclusion=self.edge_exclusion,
        )


@dataclass(frozen=True)
class ScenarioConfig:
    name: str
    surface: SurfaceParams
    polygon_kind: PolygonKind = PolygonKind.RECT
    polygon_params: Mapping[str, Any] | None = None
    grid: GridSpec = GridSpec()
    master_seed: int = 0
    model_unit: Unit = Unit.METERS

    @classmethod
    def from_dict(cls, payload: Mapping[str, Any]) -> "ScenarioConfig":
        poly_kind = PolygonKind(str(payload.get("polygon_kind", "rect")))
        grid = GridSpec(**(payload.get("grid") or {}))
        surface_raw = payload.get("surface") or {}
        # hydrate nested ripple/bump specs
        if surface_raw.get("ripples") is not None:
            surface_raw = dict(surface_raw)
            surface_raw["ripples"] = [RippleSpec(**r) for r in surface_raw.get("ripples")]
        if surface_raw.get("bump_specs") is not None:
            surface_raw = dict(surface_raw)
            surface_raw["bump_specs"] = [BumpSpec(**b) for b in surface_raw.get("bump_specs")]
        surface = SurfaceParams(**surface_raw)
        return cls(
            name=str(payload["name"]),
            surface=surface,
            polygon_kind=poly_kind,
            polygon_params=payload.get("polygon_params") or {},
            grid=grid,
            master_seed=int(payload.get("master_seed", 0)),
            model_unit=Unit(payload.get("model_unit", Unit.METERS.value)),
        )

    def to_dict(self) -> dict[str, Any]:
        out = {
            "name": self.name,
            "surface": asdict(self.surface),
            "polygon_kind": self.polygon_kind.value,
            "polygon_params": dict(self.polygon_params or {}),
            "grid": asdict(self.grid),
            "master_seed": int(self.master_seed),
            "model_unit": self.model_unit.value,
        }
        return out


@dataclass(frozen=True)
class ScenarioMaterialized:
    name: str
    master_points: pd.DataFrame
    area: SurveyArea
    grid_xy: pd.DataFrame
    grid_with_z: pd.DataFrame


class Scenario:
    def __init__(self, cfg: ScenarioConfig):
        self.cfg = cfg

    # deterministic polygon builder
    def _make_polygon(self, rng: np.random.Generator) -> Polygon:
        p = self.cfg
        if p.polygon_kind == PolygonKind.RECT:
            return make_rectangle(p.surface.width, p.surface.length)
        if p.polygon_kind == PolygonKind.ROTATED_RECT:
            base = make_rectangle(p.surface.width, p.surface.length)
            deg = float(p.polygon_params.get("rotation_deg", 0.0))
            rot = rotate_polygon(base, deg)
            return rot.intersection(base)
        if p.polygon_kind == PolygonKind.L_SHAPE:
            notch = float(p.polygon_params["notch"])
            return make_l_shape(p.surface.width, p.surface.length, notch=notch)
        if p.polygon_kind == PolygonKind.RANDOM_CONVEX:
            from .synthetic import random_convex_polygon  # reuse
            centre = tuple(p.polygon_params.get("centre", (p.surface.width/2, p.surface.length/2)))
            radius = float(p.polygon_params.get("radius", min(p.surface.width, p.surface.length)/3))
            vertices = int(p.polygon_params.get("vertices", 10))
            seed = int(p.polygon_params.get("seed", rng.integers(0, 2**31-1)))
            poly = random_convex_polygon(centre=centre, radius=radius, vertices=vertices, seed=seed)
            deg = float(p.polygon_params.get("rotation_deg", 0.0))
            return rotate_polygon(poly, deg) if deg else poly
        raise ValueError(f"Unknown polygon_kind {p.polygon_kind}")

    @staticmethod
    def _infer_half_diag(df: pd.DataFrame) -> float | None:
        xs = np.unique(df["x"].to_numpy())
        ys = np.unique(df["y"].to_numpy())
        if xs.size < 2 or ys.size < 2:
            return None
        dx = np.min(np.diff(xs)) if xs.size > 1 else None
        dy = np.min(np.diff(ys)) if ys.size > 1 else None
        if dx is None or dy is None or not np.isfinite(dx) or not np.isfinite(dy):
            return None
        return 0.5 * float(np.hypot(dx, dy))

    def materialize(self) -> ScenarioMaterialized:
        rng = numpy_rng(self.cfg.master_seed, namespace="suite.materialize")
        df = make_surface_df(self.cfg.surface)
        poly = self._make_polygon(rng)
        area = SurveyArea(polygon=poly)

        grid_xy = make_grid(area, params=self.cfg.grid.to_params())

        # auto max_dist from inferred spacing
        half_diag = self._infer_half_diag(df)
        maxd = 1.01 * half_diag if half_diag is not None else None

        nn = nearest_surface_values(df, grid_xy, max_dist=maxd)
        grid_with_z = grid_xy.copy()
        grid_with_z["z"] = nn["z_nn"]

        return ScenarioMaterialized(
            name=self.cfg.name,
            master_points=df,
            area=area,
            grid_xy=grid_xy,
            grid_with_z=grid_with_z,
        )

    def sample_surface(
        self,
        materialized: ScenarioMaterialized,
        strategy: "SamplingStrategy",
        *,
        nearest_max_dist: float | Literal["auto"] | None = "auto",
    ) -> pd.DataFrame:
        qp = strategy.sample(materialized.area, self.cfg.master_seed)
        if nearest_max_dist == "auto":
            half_diag = self._infer_half_diag(materialized.master_points)
            maxd = 1.01 * half_diag if half_diag is not None else None
        else:
            maxd = nearest_max_dist

        nn = nearest_surface_values(materialized.master_points, qp, max_dist=maxd)
        out = qp.copy()
        out["z"] = nn["z_nn"]
        return out


class SamplingStrategy:
    def sample(self, area: SurveyArea, seed: int) -> pd.DataFrame:
        raise NotImplementedError


class RandomUniform(SamplingStrategy):
    def __init__(self, n: int):
        self.n = int(n)

    def sample(self, area: SurveyArea, seed: int) -> pd.DataFrame:
        rng = numpy_rng(seed, namespace="suite.random_uniform")
        minx, miny, maxx, maxy = area.polygon.bounds
        xs: list[float] = []
        ys: list[float] = []
        while len(xs) < self.n:
            x = rng.uniform(minx, maxx)
            y = rng.uniform(miny, maxy)
            if area.polygon.contains(Point(float(x), float(y))):
                xs.append(float(x)); ys.append(float(y))
        return pd.DataFrame({"x": xs, "y": ys, "point_id": np.arange(self.n, dtype=int)})


class UniformGridWithJitter(SamplingStrategy):
    def __init__(
        self,
        spacing: float,
        jitter: float = 0.0,
        rotation_deg: float = 0.0,
        edge_exclusion: float = 0.0
    ):
        self.spacing = float(spacing)
        self.jitter = float(jitter)
        self.rotation_deg = float(rotation_deg)
        self.edge_exclusion = float(edge_exclusion)

    def sample(self, area: SurveyArea, seed: int) -> pd.DataFrame:
        base = make_grid(area, params=SamplingParams(
            grid_spacing=self.spacing,
            edge_exclusion=self.edge_exclusion,
            rotation_deg=self.rotation_deg
        ))
        if self.jitter <= 0:
            return base
        rng = numpy_rng(seed, namespace="suite.uniform_jitter")
        jx = rng.uniform(-self.jitter, self.jitter, size=len(base))
        jy = rng.uniform(-self.jitter, self.jitter, size=len(base))
        df = base.copy()
        df["x"] = df["x"] + jx
        df["y"] = df["y"] + jy
        # keep those that stay inside
        inside = [area.polygon.contains(Point(float(x), float(y))) for x, y in zip(df["x"], df["y"])]
        df = df.loc[inside].reset_index(drop=True)
        df["point_id"] = np.arange(len(df), dtype=int)
        return df


@dataclass
class ScenarioSuite:
    items: list[ScenarioConfig]

    def save_manifest(self, path: str) -> None:
        import json
        payload = {"items": [cfg.to_dict() for cfg in self.items]}
        Path(path).write_text(json.dumps(payload, indent=2), encoding="utf-8")

    @classmethod
    def load_manifest(cls, path: str) -> "ScenarioSuite":
        import json
        payload = json.loads(Path(path).read_text())
        items_raw = payload.get("items")
        if not isinstance(items_raw, Iterable):
            raise ValueError("Manifest must contain an 'items' list")
        items = [ScenarioConfig.from_dict(item) for item in items_raw]
        return cls(items)

    def names(self) -> list[str]:
        return [c.name for c in self.items]

    def get(self, name: str) -> Scenario:
        for c in self.items:
            if c.name == name:
                return Scenario(c)
        raise KeyError(name)


__all__ = [
    "ScenarioConfig",
    "ScenarioSuite",
    "Scenario",
    "ScenarioMaterialized",
    "RandomUniform",
    "UniformGridWithJitter",
    "GridSpec",
    "PolygonKind",
]



"""Lightweight geometry and sampling utilities used by the scenario suite."""
from __future__ import annotations

import math
from dataclasses import dataclass
from typing import Iterable

import numpy as np
import pandas as pd
from shapely.geometry import Point, Polygon

try:  # optional speed-up
    from scipy.spatial import cKDTree  # type: ignore

    _HAVE_SCIPY = True
except Exception:  # pragma: no cover - optional dependency missing
    _HAVE_SCIPY = False


@dataclass(slots=True)
class SurveyArea:
    polygon: Polygon

    def __post_init__(self) -> None:
        if not self.polygon.is_valid:
            raise ValueError("SurveyArea.polygon must be a valid polygon")


@dataclass(slots=True)
class SamplingParams:
    shift_x: float = 0.0
    shift_y: float = 0.0
    rotation_deg: float = 0.0
    grid_spacing: float = 3.0
    edge_exclusion: float = 1.5


def _rotate(points: np.ndarray, theta: float, origin: tuple[float, float]) -> np.ndarray:
    c, s = math.cos(theta), math.sin(theta)
    R = np.array([[c, -s], [s, c]])
    return (points - origin) @ R.T + origin


def make_grid(area: SurveyArea, params: SamplingParams) -> pd.DataFrame:
    """Axis-aligned grid optionally rotated and clipped to polygon minus buffer."""

    minx, miny, maxx, maxy = area.polygon.bounds
    xs = np.arange(minx + params.shift_x, maxx, params.grid_spacing)
    ys = np.arange(miny + params.shift_y, maxy, params.grid_spacing)
    xx, yy = np.meshgrid(xs, ys, indexing="xy")
    grid = np.column_stack([xx.ravel(), yy.ravel()])

    if params.rotation_deg:
        grid = _rotate(grid, math.radians(params.rotation_deg), area.polygon.centroid.coords[0])

    clip_poly = area.polygon.buffer(-params.edge_exclusion)
    inside = np.fromiter(
        (clip_poly.contains(Point(float(x), float(y))) for x, y in grid),
        dtype=bool,
        count=len(grid),
    )
    grid = grid[inside]

    return pd.DataFrame({
        "x": grid[:, 0],
        "y": grid[:, 1],
        "point_id": np.arange(len(grid), dtype=int),
    })


def nearest_surface_values(
    surface_df: pd.DataFrame,
    query_df: pd.DataFrame,
    *,
    max_dist: float | None = 0.10,
) -> pd.DataFrame:
    """Nearest-neighbour z lookup for query points.

    Uses SciPy cKDTree when available, otherwise falls back to a NumPy brute-force
    search (acceptable for the modest scenario grid sizes).
    """

    xyz = surface_df[["x", "y"]].to_numpy()
    qxy = query_df[["x", "y"]].to_numpy()

    if _HAVE_SCIPY:
        tree = cKDTree(xyz)
        dists, idx = tree.query(qxy, k=1)
    else:
        d2 = np.sum((qxy[:, None, :] - xyz[None, :, :]) ** 2, axis=-1)
        idx = np.argmin(d2, axis=1)
        dists = np.sqrt(d2[np.arange(len(qxy)), idx])

    z_nearest = surface_df["z"].to_numpy()[idx]
    out = query_df.copy().reset_index(drop=True)
    out["z_nn"] = z_nearest
    out["dist"] = dists

    if max_dist is not None and np.any(dists > max_dist):
        bad = np.flatnonzero(dists > max_dist)
        raise ValueError(
            f"{len(bad)} query point(s) further than {max_dist:.3f} m from the surface grid."
        )

    return out


__all__ = ["SurveyArea", "SamplingParams", "make_grid", "nearest_surface_values"]



"""Tiny unit shim to keep SurfaceParams API stable."""
from __future__ import annotations
from enum import Enum

class Unit(str, Enum):
    METERS = "m"


def convert_unit(value: float, src: Unit, dst: Unit) -> float:
    if src == dst:
        return value
    raise ValueError(f"Unsupported unit conversion {src} -> {dst}")

__all__ = ["Unit", "convert_unit"]



from survi_scenarios.suite import ScenarioSuite
from survi_scenarios.loaders import _default_suite_manifest


def test_load_suite_and_names():
    suite = ScenarioSuite.load_manifest(str(_default_suite_manifest()))
    names = suite.names()
    assert "flat_rect_small" in names
    scenario = suite.get(names[0])
    mat = scenario.materialize()
    assert not mat.master_points.empty
    assert not mat.grid_with_z.empty



from importlib import resources
from pathlib import Path
import json


def test_suite_manifest_files_exist():
    manifest = Path(resources.files("survi_scenarios.data.scenarios") / "suite_manifest.json")
    payload = json.loads(manifest.read_text())
    base = manifest.parent
    for item in payload.get("items", []):
        # manifest describes synthetic parameters; no external file per item
        assert "name" in item
    assert payload.get("items"), "suite_manifest.json should list scenarios"


def test_sdf_manifest_entries_exist():
    manifest = Path(resources.files("survi_scenarios.data.scenarios_3d") / "manifest.json")
    payload = json.loads(manifest.read_text())
    base = manifest.parent
    scenarios = payload.get("scenarios", {})
    assert scenarios, "manifest must contain scenarios"
    for name, rel in scenarios.items():
        path = (base / rel).resolve()
        assert path.exists(), f"Missing SDF scenario config for {name}: {path}"



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Callable, Optional, Mapping, Any

import numpy as np
import pandas as pd

ArrayLike = np.ndarray


@dataclass(slots=True)
class ElevationSurface:
    name: str
    samples: pd.DataFrame
    truth_z: Callable[[ArrayLike], ArrayLike]
    truth_gradient: Optional[Callable[[ArrayLike], ArrayLike]] = None
    description: str = ""
    metadata: Mapping[str, Any] = field(default_factory=dict)

    def summary(self) -> dict[str, object]:
        xy = self.samples[["x", "y"]].to_numpy()
        extent = np.max(xy, axis=0) - np.min(xy, axis=0)
        return {
            "points": int(len(self.samples)),
            "extent_xy": tuple(float(v) for v in extent),
            "description": self.description,
        }


@dataclass(slots=True)
class SDFSurface:
    name: str
    raw_surface: pd.DataFrame
    truth_phi: Callable[[ArrayLike], ArrayLike]
    truth_normals: Optional[Callable[[ArrayLike], ArrayLike]] = None
    truth_height: Optional[Callable[[ArrayLike], ArrayLike]] = None
    bounds: tuple[np.ndarray, np.ndarray] | None = None
    description: str = ""
    metadata: Mapping[str, Any] = field(default_factory=dict)

    def summary(self) -> dict[str, object]:
        coords = self.raw_surface[["x", "y", "z"]].to_numpy()
        extent = np.max(coords, axis=0) - np.min(coords, axis=0)
        return {
            "points": int(len(self.raw_surface)),
            "extent_xyz": tuple(float(v) for v in extent),
            "description": self.description,
        }


__all__ = ["ElevationSurface", "SDFSurface"]



import numpy as np
from survi_scenarios.synthetic import SurfaceParams, RippleSpec, make_surface_df


def test_surface_df_shape_and_seed():
    params = SurfaceParams(width=4.0, length=3.0, spacing=1.0, ripples=[RippleSpec(amp=0.1, wavelength=2.0)], seed=42)
    df = make_surface_df(params)
    # (length/spacing +1) * (width/spacing +1) samples
    assert len(df) == 5 * 4
    # deterministic seed
    df2 = make_surface_df(params)
    np.testing.assert_allclose(df["z"].to_numpy(), df2["z"].to_numpy())



"""Minimal RNG helper mirroring Survi's deterministic-friendly wrapper."""
from __future__ import annotations

import hashlib
import os
import numpy as np

_DEFAULT_NAMESPACE = "survi-scenarios"


def _hash_namespace(seed: int, namespace: str | None) -> int:
    label = namespace or _DEFAULT_NAMESPACE
    digest = hashlib.blake2b(f"{seed}:{label}".encode(), digest_size=8).digest()
    return int.from_bytes(digest, "big")


def base_seed() -> int:
    """Return base seed from SURVI_SEED (defaults to 0)."""

    raw = os.getenv("SURVI_SEED")
    try:
        return int(raw) if raw is not None else 0
    except (TypeError, ValueError):
        return 0


def numpy_rng(seed: int | None = None, *, namespace: str | None = None) -> np.random.Generator:
    """Return a NumPy Generator. If SURVI_DETERMINISTIC=1, derive seeds from SURVI_SEED and namespace."""

    deterministic = os.getenv("SURVI_DETERMINISTIC", "0").lower() in {"1", "true", "yes", "on"}
    if seed is not None:
        resolved = int(seed)
    elif deterministic:
        resolved = _hash_namespace(base_seed(), namespace)
    else:
        resolved = None
    return np.random.default_rng(resolved)

__all__ = ["numpy_rng", "base_seed"]



"""Synthetic 3D signed-distance field (SDF) surface factory utilities."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Iterable, Sequence

import numpy as np
import pandas as pd

from .rng import numpy_rng

__all__ = [
    "SDFNode",
    "Sphere",
    "Box",
    "Union",
    "Intersection",
    "Difference",
    "Synthetic3DSurfaceFactory",
    "Torus",
    "make_asteroid_field",
    "make_cave_network",
    "make_canal_maze",
    "make_torus",
]


def _as_point(vec: Sequence[float]) -> np.ndarray:
    arr = np.asarray(vec, dtype=float)
    if arr.shape != (3,):  # pragma: no cover - guardrail
        raise ValueError("Expected a 3D vector")
    return arr


def _normalize(vecs: np.ndarray, eps: float = 1e-9) -> np.ndarray:
    norms = np.linalg.norm(vecs, axis=1, keepdims=True)
    norms = np.clip(norms, eps, None)
    return vecs / norms


class SDFNode:
    """Abstract base class for SDF primitives and composite nodes."""

    def phi(self, points: np.ndarray) -> np.ndarray:
        raise NotImplementedError

    def gradient(self, points: np.ndarray, eps: float = 1e-4) -> np.ndarray:
        if points.size == 0:
            return np.zeros((0, 3), dtype=float)
        offsets = np.eye(3, dtype=float) * eps
        grads = []
        for offset in offsets:
            grads.append((self.phi(points + offset) - self.phi(points - offset)) / (2.0 * eps))
        return np.stack(grads, axis=1)


@dataclass(slots=True)
class Sphere(SDFNode):
    center: np.ndarray
    radius: float

    def __init__(self, *, center: Sequence[float] = (0.0, 0.0, 0.0), radius: float = 1.0) -> None:
        self.center = _as_point(center)
        self.radius = float(radius)

    def phi(self, points: np.ndarray) -> np.ndarray:  # type: ignore[override]
        diff = points - self.center
        return np.linalg.norm(diff, axis=1) - self.radius

    def gradient(self, points: np.ndarray, eps: float = 1e-4) -> np.ndarray:  # type: ignore[override]
        diff = points - self.center
        return _normalize(diff)


@dataclass(slots=True)
class Box(SDFNode):
    center: np.ndarray
    half_sizes: np.ndarray

    def __init__(self, *, center: Sequence[float] = (0.0, 0.0, 0.0), size: Sequence[float] = (1.0, 1.0, 1.0)) -> None:
        self.center = _as_point(center)
        self.half_sizes = 0.5 * np.asarray(size, dtype=float)
        if np.any(self.half_sizes <= 0):
            raise ValueError("Box size components must be positive")

    def phi(self, points: np.ndarray) -> np.ndarray:  # type: ignore[override]
        q = np.abs(points - self.center) - self.half_sizes
        outside = np.maximum(q, 0.0)
        outside_norm = np.linalg.norm(outside, axis=1)
        inside = np.minimum(np.maximum.reduce(q, axis=1), 0.0)
        return outside_norm + inside


class Union(SDFNode):
    def __init__(self, *nodes: SDFNode) -> None:
        if not nodes:
            raise ValueError("Union requires at least one node")
        self.nodes = nodes

    def phi(self, points: np.ndarray) -> np.ndarray:  # type: ignore[override]
        values = np.stack([node.phi(points) for node in self.nodes], axis=0)
        return np.min(values, axis=0)

    def gradient(self, points: np.ndarray, eps: float = 1e-4) -> np.ndarray:  # type: ignore[override]
        values = np.stack([node.phi(points) for node in self.nodes], axis=0)
        winners = np.argmin(values, axis=0)
        grads = np.stack([node.gradient(points, eps=eps) for node in self.nodes], axis=0)
        selected = np.take_along_axis(grads, winners[np.newaxis, :, np.newaxis], axis=0)[0]
        return _normalize(selected)


class Intersection(SDFNode):
    def __init__(self, *nodes: SDFNode) -> None:
        if not nodes:
            raise ValueError("Intersection requires at least one node")
        self.nodes = nodes

    def phi(self, points: np.ndarray) -> np.ndarray:  # type: ignore[override]
        values = np.stack([node.phi(points) for node in self.nodes], axis=0)
        return np.max(values, axis=0)

    def gradient(self, points: np.ndarray, eps: float = 1e-4) -> np.ndarray:  # type: ignore[override]
        values = np.stack([node.phi(points) for node in self.nodes], axis=0)
        winners = np.argmax(values, axis=0)
        grads = np.stack([node.gradient(points, eps=eps) for node in self.nodes], axis=0)
        selected = np.take_along_axis(grads, winners[np.newaxis, :, np.newaxis], axis=0)[0]
        return _normalize(selected)


class Difference(SDFNode):
    def __init__(self, a: SDFNode, b: SDFNode) -> None:
        self.a = a
        self.b = b

    def phi(self, points: np.ndarray) -> np.ndarray:  # type: ignore[override]
        pa = self.a.phi(points)
        pb = self.b.phi(points)
        return np.maximum(pa, -pb)

    def gradient(self, points: np.ndarray, eps: float = 1e-4) -> np.ndarray:  # type: ignore[override]
        pa = self.a.phi(points)
        pb = self.b.phi(points)
        grad_a = self.a.gradient(points, eps=eps)
        grad_b = -self.b.gradient(points, eps=eps)
        mask = pa >= -pb
        selected = np.where(mask[:, None], grad_a, grad_b)
        return _normalize(selected)


@dataclass(slots=True)
class Torus(SDFNode):
    """Torus aligned with the z-axis."""

    major_radius: float
    minor_radius: float
    center: np.ndarray

    def __init__(
        self,
        *,
        major_radius: float = 0.8,
        minor_radius: float = 0.25,
        center: Sequence[float] = (0.0, 0.0, 0.0),
    ) -> None:
        if major_radius <= 0 or minor_radius <= 0:
            raise ValueError("Torus radii must be positive")
        self.major_radius = float(major_radius)
        self.minor_radius = float(minor_radius)
        self.center = _as_point(center)

    def phi(self, points: np.ndarray) -> np.ndarray:  # type: ignore[override]
        shifted = points - self.center
        xy_norm = np.linalg.norm(shifted[:, :2], axis=1)
        inner = np.sqrt((xy_norm - self.major_radius) ** 2 + shifted[:, 2] ** 2)
        return inner - self.minor_radius

    def gradient(self, points: np.ndarray, eps: float = 1e-9) -> np.ndarray:  # type: ignore[override]
        shifted = points - self.center
        xy_norm = np.linalg.norm(shifted[:, :2], axis=1)
        xy_safe = np.clip(xy_norm, eps, None)
        inner = np.sqrt((xy_safe - self.major_radius) ** 2 + shifted[:, 2] ** 2)
        inner = np.clip(inner, eps, None)

        grad = np.zeros_like(shifted)
        radial_term = (xy_safe - self.major_radius) / inner
        grad[:, 0] = shifted[:, 0] * radial_term / xy_safe
        grad[:, 1] = shifted[:, 1] * radial_term / xy_safe
        grad[:, 2] = shifted[:, 2] / inner
        return _normalize(grad)


@dataclass(slots=True)
class Synthetic3DSurfaceFactory:
    root: SDFNode
    bounds: tuple[np.ndarray, np.ndarray]
    name: str = "custom"

    def __post_init__(self) -> None:
        low, high = self.bounds
        self.bounds = (_as_point(low), _as_point(high))
        if np.any(self.bounds[0] >= self.bounds[1]):
            raise ValueError("Invalid bounds: low must be strictly less than high")

    def evaluate(self, points: np.ndarray, *, with_normals: bool = False) -> tuple[np.ndarray, np.ndarray | None]:
        phi = self.root.phi(points)
        normals = None
        if with_normals:
            normals = _normalize(self.root.gradient(points))
        return phi, normals

    def sample_uniform(self, n_samples: int, *, seed: int | None = None, with_normals: bool = False) -> pd.DataFrame:
        if n_samples <= 0:
            raise ValueError("n_samples must be positive")
        rng = numpy_rng(seed, namespace="synthetic3d.uniform")
        low, high = self.bounds
        points = rng.uniform(low, high, size=(n_samples, 3))
        phi = self.root.phi(points)
        data = {
            "x": points[:, 0],
            "y": points[:, 1],
            "z": points[:, 2],
            "phi": phi,
        }
        if with_normals:
            normals = _normalize(self.root.gradient(points))
            data.update({
                "nx": normals[:, 0],
                "ny": normals[:, 1],
                "nz": normals[:, 2],
            })
        return pd.DataFrame(data)

    def sample_surface_band(
        self,
        n_samples: int,
        *,
        band_width: float = 0.15,
        oversample_factor: int = 6,
        seed: int | None = None,
        with_normals: bool = True,
        max_attempts: int = 12,
    ) -> pd.DataFrame:
        if n_samples <= 0:
            raise ValueError("n_samples must be positive")
        if band_width <= 0:
            raise ValueError("band_width must be positive")
        rng = numpy_rng(seed, namespace="synthetic3d.surface_band")
        low, high = self.bounds
        collected_pts: list[np.ndarray] = []
        collected_phi: list[np.ndarray] = []
        n_collected = 0
        attempts = 0
        while n_collected < n_samples and attempts < max_attempts:
            attempts += 1
            remaining = max(n_samples - n_collected, 1)
            batch_size = max(oversample_factor * remaining, remaining)
            batch = rng.uniform(low, high, size=(batch_size, 3))
            phi = self.root.phi(batch)
            mask = np.abs(phi) <= band_width
            if np.any(mask):
                collected_pts.append(batch[mask])
                collected_phi.append(phi[mask])
                n_collected += np.count_nonzero(mask)
        if not collected_pts or n_collected < n_samples:
            raise RuntimeError(
                "Unable to sample near-surface points; consider increasing band_width or oversample_factor"
            )
        points = np.concatenate(collected_pts, axis=0)[:n_samples]
        phi_vals = np.concatenate(collected_phi, axis=0)[:n_samples]
        data = {
            "x": points[:, 0],
            "y": points[:, 1],
            "z": points[:, 2],
            "phi": phi_vals,
        }
        if with_normals:
            normals = _normalize(self.root.gradient(points))
            data.update({
                "nx": normals[:, 0],
                "ny": normals[:, 1],
                "nz": normals[:, 2],
            })
        return pd.DataFrame(data)


def _build_sphere_union(
    rng: np.random.Generator,
    *,
    count: int,
    radius_range: tuple[float, float],
    spread: float,
) -> tuple[list[Sphere], list[Sphere]]:
    cores: list[Sphere] = []
    subtractions: list[Sphere] = []
    r_min, r_max = radius_range
    for _ in range(count):
        radius = rng.uniform(r_min, r_max)
        center = rng.normal(scale=spread, size=3)
        cores.append(Sphere(center=center, radius=radius))
        if rng.uniform() < 0.4:
            direction = rng.normal(size=3)
            direction = direction / np.linalg.norm(direction)
            crater_center = center + direction * radius * rng.uniform(0.45, 0.85)
            crater_radius = radius * rng.uniform(0.25, 0.45)
            subtractions.append(Sphere(center=crater_center, radius=crater_radius))
    return cores, subtractions


def make_asteroid_field(
    *,
    count: int = 10,
    radius_range: tuple[float, float] = (0.4, 1.2),
    spread: float = 2.5,
    seed: int | None = None,
) -> Synthetic3DSurfaceFactory:
    rng = numpy_rng(seed, namespace="synthetic3d.asteroid")
    cores, subtractions = _build_sphere_union(rng, count=count, radius_range=radius_range, spread=spread)
    core_union: SDFNode = Union(*cores) if len(cores) > 1 else cores[0]
    if subtractions:
        caverns = Union(*subtractions) if len(subtractions) > 1 else subtractions[0]
        root: SDFNode = Difference(core_union, caverns)
    else:
        root = core_union
    max_r = radius_range[1]
    bound = spread * 2.5 + max_r * 1.5
    low = np.full(3, -bound)
    high = np.full(3, bound)
    return Synthetic3DSurfaceFactory(root=root, bounds=(low, high), name="asteroid-field")


def make_cave_network(
    *,
    cavern_radius: float = 5.0,
    cavity_count: int = 12,
    tunnel_count: int = 6,
    seed: int | None = None,
) -> Synthetic3DSurfaceFactory:
    rng = numpy_rng(seed, namespace="synthetic3d.cave")
    base = Sphere(center=(0.0, 0.0, 0.0), radius=cavern_radius)
    cavities = []
    for _ in range(cavity_count):
        offset = rng.normal(scale=cavern_radius * 0.6, size=3)
        amp = rng.uniform(0.4, 0.8)
        cavities.append(Sphere(center=offset, radius=cavern_radius * 0.25 * amp))
    tunnels = []
    for _ in range(tunnel_count):
        start = rng.normal(scale=cavern_radius * 0.7, size=3)
        axis = rng.normal(size=3)
        axis = axis / np.linalg.norm(axis)
        end = start + axis * rng.uniform(cavern_radius * 0.8, cavern_radius * 1.4)
        mid = 0.5 * (start + end)
        length = np.linalg.norm(end - start)
        direction = axis
        size = np.array([cavern_radius * 0.25, cavern_radius * 0.25, length + cavern_radius * 0.4])
        rot = _orthonormal_basis(direction)
        tunnels.append(_oriented_box(center=mid, size=size, rotation=rot))
    subtractors: list[SDFNode] = []
    if cavities:
        subtractors.append(Union(*cavities) if len(cavities) > 1 else cavities[0])
    if tunnels:
        subtractors.append(Union(*tunnels) if len(tunnels) > 1 else tunnels[0])
    if subtractors:
        root: SDFNode = Difference(base, Union(*subtractors) if len(subtractors) > 1 else subtractors[0])
    else:
        root = base
    bound = cavern_radius * 1.25
    low = np.full(3, -bound)
    high = np.full(3, bound)
    return Synthetic3DSurfaceFactory(root=root, bounds=(low, high), name="cave-network")


def make_canal_maze(
    *,
    span: float = 8.0,
    height: float = 2.5,
    channel_width: float = 0.9,
    grid: int = 4,
    seed: int | None = None,
) -> Synthetic3DSurfaceFactory:
    rng = numpy_rng(seed, namespace="synthetic3d.canal")
    base = Box(center=(0.0, 0.0, 0.0), size=(span * 2, span * 2, height))
    channels: list[SDFNode] = []
    step = (2 * span) / grid
    for i in range(grid + 1):
        offset = -span + i * step
        channels.append(Box(center=(offset, 0.0, 0.0), size=(channel_width, span * 2.2, height * 0.9)))
        channels.append(Box(center=(0.0, offset, 0.0), size=(span * 2.2, channel_width, height * 0.9)))
    for _ in range(grid // 2):
        centre = rng.normal(scale=span * 0.5, size=2)
        channels.append(Box(center=(centre[0], centre[1], 0.0), size=(channel_width * 0.75, channel_width * 0.75, height * 1.1)))
    subtractor = Union(*channels)
    root: SDFNode = Difference(base, subtractor)
    bound = span * 1.05
    low = np.array([-bound, -bound, -height * 0.6])
    high = np.array([bound, bound, height * 0.6])
    return Synthetic3DSurfaceFactory(root=root, bounds=(low, high), name="canal-maze")


def make_torus(
    *,
    major_radius: float = 0.8,
    minor_radius: float = 0.25,
    bounds: tuple[Sequence[float], Sequence[float]] | None = None,
    padding: float = 0.4,
) -> Synthetic3DSurfaceFactory:
    """Convenience factory for a z-aligned torus."""

    root = Torus(major_radius=major_radius, minor_radius=minor_radius)
    if bounds is None:
        span_xy = major_radius + minor_radius + padding
        span_z = minor_radius + padding
        low = np.array([-span_xy, -span_xy, -span_z], dtype=float)
        high = np.array([span_xy, span_xy, span_z], dtype=float)
    else:
        low = _as_point(bounds[0])
        high = _as_point(bounds[1])
    return Synthetic3DSurfaceFactory(root=root, bounds=(low, high), name="torus")


def _orthonormal_basis(direction: np.ndarray) -> np.ndarray:
    direction = direction / np.linalg.norm(direction)
    if abs(direction[0]) < 0.5:
        other = np.array([1.0, 0.0, 0.0])
    else:
        other = np.array([0.0, 1.0, 0.0])
    v1 = other - np.dot(other, direction) * direction
    v1 /= np.linalg.norm(v1)
    v2 = np.cross(direction, v1)
    return np.stack((v1, v2, direction), axis=1)


def _oriented_box(*, center: np.ndarray, size: np.ndarray, rotation: np.ndarray) -> SDFNode:
    local_box = Box(center=(0.0, 0.0, 0.0), size=size)

    class _TransformedBox(SDFNode):
        def phi(self, points: np.ndarray) -> np.ndarray:  # type: ignore[override]
            local = (points - center) @ rotation
            return local_box.phi(local)

        def gradient(self, points: np.ndarray, eps: float = 1e-4) -> np.ndarray:  # type: ignore[override]
            local = (points - center) @ rotation
            grads = local_box.gradient(local, eps=eps) @ rotation.T
            return _normalize(grads)

    return _TransformedBox()



# Synthetic 2.5D surface generators extracted from Survi
from __future__ import annotations
from dataclasses import dataclass, fields
import math
import numpy as np
import pandas as pd
from shapely.affinity import rotate as _rotate_poly
from shapely.geometry import Polygon
from shapely.ops import unary_union

from .units import Unit, convert_unit
from .rng import numpy_rng


# ────────────────────────────────────────────────────────────────────────────
#  New components
# ────────────────────────────────────────────────────────────────────────────

@dataclass(slots=True)
class RippleSpec:
    """One sinusoidal ripple component."""
    amp: float = 0.0               # meters
    angle_deg: float = 0.0         # CCW from +x axis
    wavelength: float | None = None  # meters (preferred)
    cycles_across_diag: float | None = None  # dimensionless alternative

@dataclass(slots=True)
class BumpSpec:
    """One Gaussian bump or hollow."""
    amp: float                      # meters, sign controls bump (+) or hollow (-)
    radius: float                   # meters (1 sigma)
    center: tuple[float, float] | None = None  # (x,y). If None, RNG picks.

# ────────────────────────────────────────────────────────────────────────────
#  Surface synthesis
# ────────────────────────────────────────────────────────────────────────────

@dataclass(slots=True)
class SurfaceParams:
    width: float = 10.0
    length: float = 10.0
    spacing: float = 0.10
    slope_mag: float = 0.0
    slope_angle_deg: float = 0.0
    # legacy single ripple (kept for back compat)
    ripple_amp: float = 0.0
    ripple_freq: float = 1.0
    # legacy uniform bumps (kept for back compat)
    bumps: int = 0
    bump_amp: float = 0.0
    bump_radius: float = 0.5
    # new flexible components
    ripples: list[RippleSpec] | None = None
    bump_specs: list[BumpSpec] | None = None

    noise_mean: float = 0.0
    noise_std: float = 0.0
    seed: int | None = None
    unit: Unit = Unit.METERS

    def __post_init__(self):
        if self.unit == Unit.METERS:
            return
        cf = convert_unit(1.0, self.unit, Unit.METERS)  # to meters

        distance_fields = {
            "width", "length", "spacing", "ripple_amp", "bump_amp",
            "bump_radius", "noise_mean", "noise_std"
        }
        inverse_distance_fields = {"ripple_freq"}  # cycles per diag length proxy

        for f in fields(self):
            name = f.name; val = getattr(self, name)
            if val is None:
                continue
            if name in distance_fields:
                setattr(self, name, val * cf)
            elif name in inverse_distance_fields:
                setattr(self, name, val / cf)

        # nested conversions
        if self.ripples:
            for r in self.ripples:
                r.amp *= cf
                if r.wavelength is not None:
                    r.wavelength *= cf
        if self.bump_specs:
            for b in self.bump_specs:
                b.amp *= cf
                b.radius *= cf
                if b.center is not None:
                    b.center = (b.center[0] * cf, b.center[1] * cf)

        self.unit = Unit.METERS

def make_surface_df(p: SurfaceParams) -> pd.DataFrame:
    """Return a validated-like DataFrame with (x,y,z)."""
    rng = numpy_rng(p.seed, namespace="synthetic.surface")

    nx = int(math.floor(p.width / p.spacing)) + 1
    ny = int(math.floor(p.length / p.spacing)) + 1

    xs = np.linspace(0, p.width, nx)
    ys = np.linspace(0, p.length, ny)
    X, Y = np.meshgrid(xs, ys, indexing="xy")

    # slope
    theta = math.radians(p.slope_angle_deg)
    Z = p.slope_mag * (X * math.cos(theta) + Y * math.sin(theta))

    # new: multiple ripples
    if p.ripples:
        diag = math.hypot(p.width, p.length)
        for r in p.ripples:
            ang = math.radians(r.angle_deg)
            s = X * math.cos(ang) + Y * math.sin(ang)  # coordinate along direction
            if r.wavelength and r.wavelength > 0:
                Z += r.amp * np.sin(2 * math.pi * s / r.wavelength)
            elif r.cycles_across_diag and diag > 0:
                wl = diag / r.cycles_across_diag
                Z += r.amp * np.sin(2 * math.pi * s / wl)
    else:
        # legacy single ripple (unchanged semantics)
        if p.ripple_amp:
            diag = math.hypot(p.width, p.length)
            if diag > 0:
                Z += p.ripple_amp * np.sin(2 * math.pi * p.ripple_freq * (X + Y) / diag)

    # new: flexible bumps
    if p.bump_specs:
        for b in p.bump_specs:
            if b.center is None:
                cx = rng.uniform(0, p.width)
                cy = rng.uniform(0, p.length)
            else:
                cx, cy = b.center
            r2 = (X - cx) ** 2 + (Y - cy) ** 2
            Z += b.amp * np.exp(-r2 / (b.radius ** 2))
    else:
        # legacy uniform bumps
        for _ in range(p.bumps):
            cx = rng.uniform(0, p.width)
            cy = rng.uniform(0, p.length)
            amp = rng.uniform(-p.bump_amp, p.bump_amp)
            r2 = (X - cx) ** 2 + (Y - cy) ** 2
            Z += amp * np.exp(-r2 / (p.bump_radius ** 2))

    if p.noise_std > 0:
        Z += rng.normal(loc=p.noise_mean, scale=p.noise_std, size=Z.shape)

    return pd.DataFrame({"x": X.ravel(), "y": Y.ravel(), "z": Z.ravel()})

# polygon helpers (unchanged)
def make_rectangle(width: float, length: float) -> Polygon:
    return Polygon([(0, 0), (width, 0), (width, length), (0, length)])

def make_l_shape(width: float, length: float, notch: float) -> Polygon:
    if notch <= 0 or notch >= min(width, length):
        raise ValueError("`notch` must be positive and smaller than both sides")
    outer = Polygon([(0, 0), (width, 0), (width, length), (0, length)])
    notch_poly = Polygon([(width-notch, 0), (width, 0), (width, notch), (width-notch, notch)])
    poly = unary_union(outer.difference(notch_poly))
    if not poly.is_valid or poly.geom_type != "Polygon":
        raise RuntimeError("L-shape generation failed")
    return poly

def random_convex_polygon(
    centre: tuple[float, float],
    radius: float,
    vertices: int = 8,
    angle_jitter: float = 0.15,
    seed: int | None = None,
) -> Polygon:
    """
    Generate a random (roughly) convex polygon around `centre` with
    `vertices` points and average radius `radius`.
    """
    rng = numpy_rng(seed, namespace="synthetic.random_polygon")
    angles = np.linspace(0, 2 * math.pi, vertices, endpoint=False)
    angles += rng.uniform(-angle_jitter, angle_jitter, size=vertices)
    rng.shuffle(angles)

    pts = [
        (
            centre[0] + radius * rng.uniform(0.85, 1.15) * math.cos(a),
            centre[1] + radius * rng.uniform(0.85, 1.15) * math.sin(a),
        )
        for a in angles
    ]
    poly = Polygon(pts).convex_hull
    if not poly.is_valid:
        poly = poly.buffer(0)  # fix self-intersections if any
    return poly

def rotate_polygon(poly: Polygon, deg: float, origin: tuple[float, float] | str = "centroid") -> Polygon:
    return _rotate_poly(poly, deg, origin=origin, use_radians=False)

__all__ = [
    "SurfaceParams",
    "RippleSpec",
    "BumpSpec",
    "make_surface_df",
    "make_rectangle",
    "make_l_shape",
    "random_convex_polygon",
    "rotate_polygon",
]



# ABOUTME: Loader for trenchfoot trench mesh scenarios as SDF surfaces.
# ABOUTME: Provides list/load functions for trenchfoot package integration.
"""Trenchfoot scenario loader integration.

Loads trench mesh scenarios from the trenchfoot package as SDFSurface objects.
Requires the [trenchfoot] optional dependency.
"""
from __future__ import annotations

from pathlib import Path
from typing import Callable

import numpy as np
import pandas as pd

from .types import SDFSurface

_trenchfoot = None
_trimesh = None
_trenchfoot_import_error: Exception | None = None
_trimesh_import_error: Exception | None = None


def _ensure_trenchfoot() -> None:
    """Lazy import trenchfoot and trimesh."""
    global _trenchfoot, _trimesh, _trenchfoot_import_error, _trimesh_import_error

    if _trenchfoot_import_error is not None:
        raise _trenchfoot_import_error
    if _trimesh_import_error is not None:
        raise _trimesh_import_error

    if _trenchfoot is None:
        try:
            import trenchfoot as tf
            _trenchfoot = tf
        except ImportError as exc:
            err = ImportError(
                "trenchfoot is required for trench scenarios. "
                "Install with: pip install surface-scenarios[trenchfoot]"
            )
            err.__cause__ = exc
            _trenchfoot_import_error = err
            raise err

    if _trimesh is None:
        try:
            import trimesh
            _trimesh = trimesh
        except ImportError as exc:
            err = ImportError(
                "trimesh is required for trench scenarios. "
                "Install with: pip install surface-scenarios[trenchfoot]"
            )
            err.__cause__ = exc
            _trimesh_import_error = err
            raise err


def _get_scenarios_path() -> Path:
    """Get the path to bundled trenchfoot scenarios."""
    _ensure_trenchfoot()
    from importlib import resources
    # trenchfoot bundles scenarios in its package
    return Path(resources.files("trenchfoot") / "scenarios")


def list_trenchfoot_scenarios() -> list[str]:
    """List available trenchfoot scenario names.

    Returns:
        List of scenario names (e.g., ['S01_straight_vwalls', ...])

    Raises:
        ImportError: If trenchfoot is not installed.
    """
    _ensure_trenchfoot()
    scenarios_path = _get_scenarios_path()
    if not scenarios_path.exists():
        return []

    names = []
    for child in scenarios_path.iterdir():
        if child.is_dir() and (child / "trench_scene.obj").exists():
            names.append(child.name)
    return sorted(names)


def load_trenchfoot_scenario(
    name: str,
    *,
    mesh_variant: str = "trench_scene.obj",
) -> SDFSurface:
    """Load a trenchfoot scenario as an SDFSurface.

    The trenchfoot mesh is loaded via trimesh, and SDF values are computed
    using trimesh's proximity queries.

    Args:
        name: Scenario name (e.g., 'S01_straight_vwalls')
        mesh_variant: Which mesh file to load. Options:
            - 'trench_scene.obj' (default): Full scene mesh
            - 'meshes/trench_scene_culled.obj': Culled mesh variant

    Returns:
        SDFSurface with the mesh vertices as raw_surface and SDF truth functions.

    Raises:
        ImportError: If trenchfoot or trimesh is not installed.
        ValueError: If the scenario or mesh file is not found.
    """
    _ensure_trenchfoot()
    scenarios_path = _get_scenarios_path()
    scenario_dir = scenarios_path / name

    if not scenario_dir.exists():
        available = list_trenchfoot_scenarios()
        raise ValueError(
            f"Unknown trenchfoot scenario '{name}'. Available: {available}"
        )

    mesh_path = scenario_dir / mesh_variant
    if not mesh_path.exists():
        raise ValueError(
            f"Mesh file not found: {mesh_path}. "
            f"Available variants: trench_scene.obj, meshes/trench_scene_culled.obj"
        )

    mesh = _trimesh.load(str(mesh_path), force="mesh")

    # Ensure normals are consistent
    mesh.fix_normals()

    # Build surface dataframe from mesh vertices
    vertices = mesh.vertices
    vertex_normals = mesh.vertex_normals

    surface_df = pd.DataFrame({
        "x": vertices[:, 0],
        "y": vertices[:, 1],
        "z": vertices[:, 2],
        "nx": vertex_normals[:, 0],
        "ny": vertex_normals[:, 1],
        "nz": vertex_normals[:, 2],
        "is_surface": True,
        "source": f"trenchfoot:{name}",
    })

    # Compute bounds with padding
    padding = 0.15
    coords = vertices
    low = np.min(coords, axis=0) - padding
    high = np.max(coords, axis=0) + padding

    # Build SDF truth functions using trimesh proximity
    def _phi_fn(points: np.ndarray) -> np.ndarray:
        """Compute signed distance to mesh surface."""
        pts = np.asarray(points, dtype=float)
        closest_pts, distances, _ = _trimesh.proximity.closest_point(mesh, pts)
        distances = np.asarray(distances, dtype=float)

        # Determine sign based on whether points are inside or outside the mesh
        # Use ray casting to determine inside/outside
        contains = mesh.contains(pts)
        sign = np.where(contains, -1.0, 1.0)

        return distances * sign

    def _normals_fn(points: np.ndarray) -> np.ndarray:
        """Compute surface normals at closest points."""
        pts = np.asarray(points, dtype=float)
        closest_pts, _, tri_idx = _trimesh.proximity.closest_point(mesh, pts)
        tri_idx = np.asarray(tri_idx, dtype=int)

        # Get face normals at closest triangles
        normals = mesh.face_normals[np.clip(tri_idx, 0, mesh.face_normals.shape[0] - 1)].copy()

        # Flip normals for points inside the mesh
        delta = pts - closest_pts
        flip = np.sum(delta * normals, axis=1) < 0.0
        normals[flip] *= -1.0

        # Handle missing triangles
        missing = tri_idx < 0
        if np.any(missing):
            fallback = delta[missing]
            fallback_norm = np.linalg.norm(fallback, axis=1, keepdims=True)
            normals[missing] = fallback / np.clip(fallback_norm, 1e-9, None)

        return normals

    # Load metadata from scene.json if available
    metadata: dict = {
        "source": "trenchfoot",
        "scenario": name,
        "mesh_variant": mesh_variant,
        "mesh_path": str(mesh_path),
        "vertex_count": len(vertices),
        "face_count": len(mesh.faces),
    }

    scene_json = scenario_dir / "scene.json"
    if scene_json.exists():
        import json
        try:
            with scene_json.open("r", encoding="utf-8") as fh:
                spec = json.load(fh)
            metadata["scene_spec"] = spec
        except Exception:
            pass

    metrics_json = scenario_dir / "metrics.json"
    if metrics_json.exists():
        import json
        try:
            with metrics_json.open("r", encoding="utf-8") as fh:
                metrics = json.load(fh)
            metadata["metrics"] = metrics
        except Exception:
            pass

    return SDFSurface(
        name=name,
        raw_surface=surface_df,
        truth_phi=_phi_fn,
        truth_normals=_normals_fn,
        bounds=(low, high),
        description=f"Trenchfoot scenario '{name}' loaded as SDF surface.",
        metadata=metadata,
    )


__all__ = ["list_trenchfoot_scenarios", "load_trenchfoot_scenario"]



"""Public scenario loader API for the standalone library."""
from __future__ import annotations

import os
from dataclasses import dataclass, replace
from dataclasses import asdict
import json
from functools import lru_cache
from importlib import resources
from pathlib import Path
from typing import Callable, Iterable, Mapping, MutableMapping

import numpy as np
import pandas as pd

try:  # optional mesh support
    import trimesh  # type: ignore
except ModuleNotFoundError:  # pragma: no cover - optional
    trimesh = None  # type: ignore[assignment]

from .synthetic import RippleSpec, SurfaceParams, make_surface_df
from .synthetic_3d import (
    Synthetic3DSurfaceFactory,
    make_asteroid_field,
    make_canal_maze,
    make_cave_network,
    make_torus,
)
from .suite import ScenarioSuite, Scenario
from .types import ElevationSurface, SDFSurface


FactoryBuilder = Callable[..., Synthetic3DSurfaceFactory]

_FACTORIES: dict[str, FactoryBuilder] = {
    "torus": make_torus,
    "asteroid_field": make_asteroid_field,
    "cave_network": make_cave_network,
    "canal_maze": make_canal_maze,
}

_HEIGHTFIELD_SUBDIV = 2
_HEIGHTFIELD_PADDING = 0.15


# ── Elevation scenarios ------------------------------------------------------

def _default_suite_manifest() -> Path:
    env = os.getenv("SURVI_SCENARIO_MANIFEST")
    if env:
        return Path(env)
    return Path(resources.files("survi_scenarios.data.scenarios") / "suite_manifest.json")


@lru_cache(maxsize=None)
def _load_suite(manifest_path: str) -> ScenarioSuite:
    return ScenarioSuite.load_manifest(manifest_path)


def list_elevation_scenarios(manifest_path: str | Path | None = None) -> list[str]:
    manifest = Path(manifest_path) if manifest_path is not None else _default_suite_manifest()
    if not manifest.exists():
        raise FileNotFoundError(f"Scenario manifest not found at {manifest}")
    suite = _load_suite(str(manifest.resolve()))
    return suite.names()


def _build_truth_fn(clean_df: pd.DataFrame) -> Callable[[np.ndarray], np.ndarray]:
    xs = np.unique(clean_df["x"].to_numpy())
    ys = np.unique(clean_df["y"].to_numpy())
    grid = (
        clean_df.pivot(index="y", columns="x", values="z")
        .reindex(index=ys, columns=xs)
        .to_numpy()
    )
    x_step = xs[1] - xs[0] if len(xs) > 1 else 1.0
    y_step = ys[1] - ys[0] if len(ys) > 1 else 1.0

    def truth_fn(points: np.ndarray) -> np.ndarray:
        pts = np.asarray(points, dtype=float)
        xp = np.clip((pts[:, 0] - xs[0]) / x_step, 0.0, len(xs) - 1.000001)
        yp = np.clip((pts[:, 1] - ys[0]) / y_step, 0.0, len(ys) - 1.000001)
        x0 = np.floor(xp).astype(int)
        y0 = np.floor(yp).astype(int)
        x1 = np.clip(x0 + 1, 0, len(xs) - 1)
        y1 = np.clip(y0 + 1, 0, len(ys) - 1)
        tx = xp - x0
        ty = yp - y0
        z00 = grid[y0, x0]
        z01 = grid[y1, x0]
        z10 = grid[y0, x1]
        z11 = grid[y1, x1]
        z0 = (1 - tx) * z00 + tx * z10
        z1 = (1 - tx) * z01 + tx * z11
        return (1 - ty) * z0 + ty * z1

    return truth_fn


def load_elevation_scenario(
    name: str,
    *,
    manifest_path: str | Path | None = None,
) -> ElevationSurface:
    manifest = Path(manifest_path) if manifest_path is not None else _default_suite_manifest()
    if not manifest.exists():
        raise FileNotFoundError(f"Scenario manifest not found at {manifest}")

    suite = _load_suite(str(manifest.resolve()))
    scenario = None
    try:
        scenario = suite.get(name)
    except KeyError:
        lowered = name.lower()
        for cfg in suite.items:
            if cfg.name.lower() == lowered:
                scenario = Scenario(cfg)
                break
        if scenario is None:
            raise KeyError(name) from None

    params = scenario.cfg.surface
    df = make_surface_df(params)

    params_clean = replace(params, noise_std=0.0, noise_mean=0.0)
    df_clean = make_surface_df(params_clean)

    if len(np.unique(df_clean["x"])) < 2 or len(np.unique(df_clean["y"])) < 2:
        raise ValueError(
            f"Scenario '{name}' does not provide a gridded surface; unable to build analytic interpolation."
        )

    truth_fn = _build_truth_fn(df_clean)

    metadata = {
        "scenario": scenario.cfg.name,
        "polygon_kind": scenario.cfg.polygon_kind.value,
        "grid": asdict(scenario.cfg.grid),
        "surface_params": asdict(scenario.cfg.surface),
        "model_unit": scenario.cfg.model_unit.value,
        "manifest": str(manifest.resolve()),
    }

    return ElevationSurface(
        name=scenario.cfg.name,
        samples=df,
        truth_z=truth_fn,
        description=f"Scenario '{scenario.cfg.name}' from the suite manifest.",
        metadata=metadata,
    )


# ── SDF scenarios -----------------------------------------------------------

@dataclass(slots=True)
class _SDFScenarioConfig:
    name: str
    factory_type: str
    factory_params: Mapping[str, object]
    description: str
    sampling: Mapping[str, object]
    metadata: MutableMapping[str, object]
    manifest_path: Path
    config_path: Path


def _default_sdf_manifest() -> Path:
    env = os.getenv("SURVI_SDF_MANIFEST")
    if env:
        return Path(env)
    return Path(resources.files("survi_scenarios.data.scenarios_3d") / "manifest.json")


@lru_cache(maxsize=None)
def _load_sdf_manifest(path: str) -> dict[str, Path]:
    manifest_path = Path(path)
    with manifest_path.open("r", encoding="utf-8") as fh:
        payload = json.load(fh)
    scenarios = payload.get("scenarios")
    if scenarios is None:
        raise ValueError(f"Manifest {manifest_path} missing 'scenarios' section")

    base = manifest_path.parent
    mapping: dict[str, Path] = {}
    if isinstance(scenarios, dict):
        items = scenarios.items()
    elif isinstance(scenarios, Iterable):
        items = []
        for item in scenarios:
            if not isinstance(item, dict) or "name" not in item:
                raise ValueError(f"Manifest entry must be a dict with a 'name': {item}")
            rel = item.get("path") or f"{item['name']}.json"
            items.append((item["name"], rel))
    else:
        raise ValueError(f"Unsupported scenarios payload in manifest {manifest_path}")

    for name, rel_path in items:
        resolved = (base / rel_path).resolve()
        mapping[str(name)] = resolved
    return mapping


def list_sdf_scenarios(manifest_path: str | Path | None = None) -> list[str]:
    manifest = Path(manifest_path) if manifest_path is not None else _default_sdf_manifest()
    names: set[str] = set()
    try:
        mapping = _load_sdf_manifest(str(manifest.resolve()))
        names.update(mapping.keys())
    except (FileNotFoundError, ValueError):
        pass

    try:
        elevation_names = list_elevation_scenarios(manifest_path)
        names.update(elevation_names)
    except (FileNotFoundError, ValueError):
        if manifest_path is None:
            try:
                names.update(list_elevation_scenarios(None))
            except Exception:
                pass
    return sorted(names)


def _load_config(name: str, *, manifest_path: str | Path | None) -> _SDFScenarioConfig:
    manifest = Path(manifest_path) if manifest_path is not None else _default_sdf_manifest()
    mapping = _load_sdf_manifest(str(manifest.resolve()))
    try:
        cfg_path = mapping[name]
    except KeyError as exc:
        available = ", ".join(sorted(mapping))
        raise KeyError(f"Unknown SDF scenario '{name}'. Available: {available}") from exc

    with cfg_path.open("r", encoding="utf-8") as fh:
        payload = json.load(fh)

    factory_section = payload.get("factory")
    if not isinstance(factory_section, dict) or "type" not in factory_section:
        raise ValueError(f"Scenario {cfg_path} must declare a 'factory' with a 'type'")
    factory_type = str(factory_section["type"]).lower()
    factory_params = dict(factory_section.get("params") or {})

    sampling_section = payload.get("sampling") or {}
    if not isinstance(sampling_section, Mapping):
        raise ValueError(f"Scenario {cfg_path} has invalid sampling configuration: {sampling_section}")

    metadata = dict(payload.get("metadata") or {})
    description = str(payload.get("description") or f"Scenario '{name}'")
    metadata.setdefault("config_path", str(cfg_path))
    metadata.setdefault("manifest_path", str(manifest.resolve()))
    metadata.setdefault("factory_type", factory_type)

    return _SDFScenarioConfig(
        name=str(payload.get("name") or name),
        factory_type=factory_type,
        factory_params=factory_params,
        description=description,
        sampling=sampling_section,
        metadata=metadata,
        manifest_path=manifest.resolve(),
        config_path=cfg_path,
    )


def _build_factory(cfg: _SDFScenarioConfig) -> Synthetic3DSurfaceFactory:
    builder = _FACTORIES.get(cfg.factory_type)
    if builder is None:
        available = ", ".join(sorted(_FACTORIES))
        raise ValueError(f"Unsupported factory '{cfg.factory_type}'. Options: {available}")
    return builder(**cfg.factory_params)


def _sample_surface(
    factory: Synthetic3DSurfaceFactory,
    sampling: Mapping[str, object],
    *,
    seed: int,
) -> pd.DataFrame:
    strategy = str(sampling.get("strategy", "surface_band")).lower()
    draws = int(sampling.get("n_samples", 4096))
    local_seed = int(sampling.get("seed", seed))

    if strategy == "surface_band":
        band_width = float(sampling.get("band_width", 0.08))
        oversample = int(sampling.get("oversample_factor", 10))
        max_attempts = int(sampling.get("max_attempts", 12))
        with_normals = bool(sampling.get("with_normals", True))
        df = factory.sample_surface_band(
            draws,
            band_width=band_width,
            oversample_factor=oversample,
            seed=local_seed,
            with_normals=with_normals,
            max_attempts=max_attempts,
        )
    elif strategy == "uniform":
        with_normals = bool(sampling.get("with_normals", True))
        df = factory.sample_uniform(
            draws,
            seed=local_seed,
            with_normals=with_normals,
        )
    else:
        raise ValueError(f"Unknown sampling strategy '{strategy}'")

    return df.reset_index(drop=True)


def load_sdf_scenario(
    name: str,
    *,
    seed: int = 0,
    manifest_path: str | Path | None = None,
) -> SDFSurface:
    try:
        cfg = _load_config(name, manifest_path=manifest_path)
    except (FileNotFoundError, KeyError, ValueError):
        return _load_heightfield_surface(name, seed=seed, manifest_path=manifest_path)

    factory = _build_factory(cfg)
    samples = _sample_surface(factory, cfg.sampling, seed=seed)

    points = samples[["x", "y", "z"]].to_numpy(dtype=float)
    phi, normals = factory.evaluate(points, with_normals=True)

    enriched = samples.copy()
    enriched["phi"] = phi
    if "nx" not in enriched.columns or "ny" not in enriched.columns or "nz" not in enriched.columns:
        normals = normals if normals is not None else factory.root.gradient(points)
        enriched["nx"] = normals[:, 0]
        enriched["ny"] = normals[:, 1]
        enriched["nz"] = normals[:, 2]
    enriched["is_surface"] = True
    enriched["source"] = cfg.name

    low, high = factory.bounds

    def _phi_fn(query: np.ndarray) -> np.ndarray:
        pts = np.asarray(query, dtype=float)
        return factory.root.phi(pts)

    def _normals_fn(query: np.ndarray) -> np.ndarray:
        pts = np.asarray(query, dtype=float)
        return factory.root.gradient(pts)

    metadata = dict(cfg.metadata)
    metadata.setdefault("samples", len(enriched))
    metadata.setdefault("seed", int(seed))

    return SDFSurface(
        name=cfg.name,
        raw_surface=enriched,
        truth_phi=_phi_fn,
        truth_normals=_normals_fn,
        bounds=(low, high),
        description=cfg.description,
        metadata=metadata,
    )


def _heightfield_to_mesh(df: pd.DataFrame) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    grid = (
        df.pivot_table(index="y", columns="x", values="z", aggfunc="mean")
        .sort_index()
        .sort_index(axis=1)
    )
    xs = grid.columns.to_numpy(dtype=float)
    ys = grid.index.to_numpy(dtype=float)
    Z = grid.to_numpy(dtype=float)
    ny, nx = Z.shape
    if nx < 2 or ny < 2:
        raise ValueError("Heightfield requires at least a 2x2 grid to form a mesh")

    if _HEIGHTFIELD_SUBDIV > 1:
        xs_dense = np.linspace(xs[0], xs[-1], (len(xs) - 1) * _HEIGHTFIELD_SUBDIV + 1)
        ys_dense = np.linspace(ys[0], ys[-1], (len(ys) - 1) * _HEIGHTFIELD_SUBDIV + 1)

        Z_interp_x = np.empty((ny, len(xs_dense)), dtype=float)
        for row_idx in range(ny):
            Z_interp_x[row_idx] = np.interp(xs_dense, xs, Z[row_idx])

        Z_dense = np.empty((len(ys_dense), len(xs_dense)), dtype=float)
        for col_idx in range(len(xs_dense)):
            Z_dense[:, col_idx] = np.interp(ys_dense, ys, Z_interp_x[:, col_idx])
    else:
        xs_dense = xs
        ys_dense = ys
        Z_dense = Z

    X, Y = np.meshgrid(xs_dense, ys_dense, indexing="xy")
    vertices = np.column_stack((X.ravel(), Y.ravel(), Z_dense.ravel()))
    ny_dense, nx_dense = Z_dense.shape

    faces: list[list[int]] = []
    for j in range(ny_dense - 1):
        row = j * nx_dense
        next_row = (j + 1) * nx_dense
        for i in range(nx_dense - 1):
            idx00 = row + i
            idx10 = row + i + 1
            idx01 = next_row + i
            idx11 = next_row + i + 1
            faces.append([idx00, idx10, idx11])
            faces.append([idx00, idx11, idx01])
    return vertices, np.asarray(faces, dtype=np.int64), xs_dense, ys_dense


def _load_heightfield_surface(
    name: str,
    *,
    seed: int,
    manifest_path: str | Path | None,
) -> SDFSurface:
    _ensure_trimesh()
    try:
        dataset = load_elevation_scenario(name, manifest_path=manifest_path)
    except FileNotFoundError as exc:
        raise ValueError(f"Scenario manifest not found at {manifest_path}") from exc
    except (KeyError, ValueError) as exc:
        try:
            available = list_elevation_scenarios(manifest_path)
        except Exception:
            available = list_elevation_scenarios(None)
        raise ValueError(
            f"Unknown elevation scenario '{name}'. Options: {sorted(available)}"
        ) from exc

    vertices, faces, xs_dense, ys_dense = _heightfield_to_mesh(dataset.samples)
    mesh = trimesh.Trimesh(vertices=vertices, faces=faces, process=False)
    mesh.update_faces(mesh.unique_faces())
    mesh.remove_unreferenced_vertices()
    mesh.fix_normals()
    if mesh.face_normals.shape[0] and np.mean(mesh.face_normals[:, 2]) < 0.0:
        mesh.invert()

    vertex_normals = mesh.vertex_normals
    surface_df = pd.DataFrame(mesh.vertices, columns=["x", "y", "z"])
    surface_df["nx"] = vertex_normals[:, 0]
    surface_df["ny"] = vertex_normals[:, 1]
    surface_df["nz"] = vertex_normals[:, 2]
    surface_df["is_surface"] = True
    surface_df["source"] = f"{dataset.name}_heightfield_mesh"

    coords = surface_df[["x", "y", "z"]].to_numpy(dtype=float)
    padding = _HEIGHTFIELD_PADDING
    low = np.min(coords, axis=0) - padding
    high = np.max(coords, axis=0) + padding

    def _phi_fn(points: np.ndarray) -> np.ndarray:
        pts = np.asarray(points, dtype=float)
        closest_pts, distances, _ = trimesh.proximity.closest_point(mesh, pts)
        distances = np.asarray(distances, dtype=float)
        heights = dataset.truth_z(pts[:, :2])
        sign = np.where(pts[:, 2] >= heights, 1.0, -1.0)
        return distances * sign

    def _truth_height_fn(points: np.ndarray) -> np.ndarray:
        pts = np.asarray(points, dtype=float)
        if pts.ndim != 2:
            raise ValueError("Heightfield truth expects an (N, D) array")
        if pts.shape[1] == 3:
            xy = pts[:, :2]
        elif pts.shape[1] == 2:
            xy = pts
        else:
            raise ValueError("Heightfield truth expects 2D or 3D points")
        return dataset.truth_z(xy)

    def _normals_fn(points: np.ndarray) -> np.ndarray:
        pts = np.asarray(points, dtype=float)
        closest_pts, _, tri_idx = trimesh.proximity.closest_point(mesh, pts)
        tri_idx = np.asarray(tri_idx, dtype=int)
        normals = mesh.face_normals[np.clip(tri_idx, 0, mesh.face_normals.shape[0] - 1)].copy()
        delta = pts - closest_pts
        flip = np.sum(delta * normals, axis=1) < 0.0
        normals[flip] *= -1.0
        missing = tri_idx < 0
        if np.any(missing):
            fallback = delta[missing]
            fallback_norm = np.linalg.norm(fallback, axis=1, keepdims=True)
            normals[missing] = fallback / np.clip(fallback_norm, 1e-9, None)
        return normals

    metadata = dict(dataset.metadata)
    metadata.update({
        "dataset_mode": "scenario_heightfield",
        "scenario": dataset.name,
        "grid_shape": (int(len(ys_dense)), int(len(xs_dense))),
        "heightfield_subdivision": int(_HEIGHTFIELD_SUBDIV),
        "padding": float(padding),
        "factory_type": "heightfield",
        "seed": int(seed),
    })

    description = dataset.description or f"Heightfield scenario '{dataset.name}' repurposed as an SDF surface."

    return SDFSurface(
        name=dataset.name,
        raw_surface=surface_df,
        truth_phi=_phi_fn,
        truth_normals=_normals_fn,
        truth_height=_truth_height_fn,
        bounds=(low, high),
        description=description,
        metadata=metadata,
    )


def _ensure_trimesh() -> None:
    if trimesh is None:
        raise RuntimeError(
            "trimesh is required to sample mesh-backed SDF scenarios. Install the optional "
            "'mesh' extra (e.g. pip install survi-scenarios[mesh]) or choose an analytic scenario."
        )


__all__ = [
    "list_elevation_scenarios",
    "load_elevation_scenario",
    "list_sdf_scenarios",
    "load_sdf_scenario",
]



from pathlib import Path

from survi_scenarios import (
    list_elevation_scenarios,
    load_elevation_scenario,
    list_sdf_scenarios,
    load_sdf_scenario,
)


def test_list_elevation_scenarios_non_empty():
    names = list_elevation_scenarios()
    assert "flat_rect_small" in names


def test_load_elevation_scenario_truth_matches_samples():
    ds = load_elevation_scenario("flat_rect_small")
    # Evaluate truth on first few points
    pts = ds.samples[["x", "y"]].to_numpy()[:5]
    truth = ds.truth_z(pts)
    # should match z (within noise tolerance, but this scenario has tiny noise)
    sample_z = ds.samples["z"].to_numpy()[:5]
    assert truth.shape == sample_z.shape


def test_list_sdf_scenarios_includes_torus():
    names = list_sdf_scenarios()
    assert "torus_compact" in names


def test_load_sdf_scenario_basic():
    sdf = load_sdf_scenario("torus_compact", seed=1)
    assert not sdf.raw_surface.empty
    assert callable(sdf.truth_phi)
    # evaluate phi on first few points
    pts = sdf.raw_surface[["x", "y", "z"]].to_numpy()[:3]
    phi = sdf.truth_phi(pts)
    assert phi.shape[0] == 3


def test_env_override_manifest(monkeypatch, tmp_path):
    # Build a minimal manifest that points to the existing data file but via tmp path
    suite_manifest = tmp_path / "suite_manifest.json"
    suite_manifest.write_text(
        """{"items": [{"name": "env_flat", "surface": {"width": 1, "length": 1, "spacing": 1, "slope_mag": 0, "slope_angle_deg": 0, "ripple_amp": 0, "ripple_freq": 1, "bumps": 0, "bump_amp": 0, "bump_radius": 1, "ripples": null, "bump_specs": null, "noise_mean": 0, "noise_std": 0, "seed": 1, "unit": "m"}, "grid": {"grid_spacing": 1, "edge_exclusion": 0, "rotation_deg": 0, "shift_x": 0, "shift_y": 0}, "master_seed": 0}] }""",
        encoding="utf-8",
    )
    monkeypatch.setenv("SURVI_SCENARIO_MANIFEST", str(suite_manifest))
    names = list_elevation_scenarios()
    assert names == ["env_flat"]
    ds = load_elevation_scenario("env_flat")
    assert ds.name == "env_flat"


def test_env_override_sdf_manifest(monkeypatch, tmp_path):
    src_manifest = (tmp_path / "manifest.json")
    src_manifest.write_text(
        """{"scenarios": {"env_torus": "torus_compact.json"}}""",
        encoding="utf-8",
    )
    # link to bundled torus file
    torus_src = Path(__file__).resolve().parents[1] / "src" / "survi_scenarios" / "data" / "scenarios_3d" / "torus_compact.json"
    (tmp_path / "torus_compact.json").write_bytes(torus_src.read_bytes())

    monkeypatch.setenv("SURVI_SDF_MANIFEST", str(src_manifest))
    names = list_sdf_scenarios()
    assert "env_torus" in names
    sdf = load_sdf_scenario("env_torus", seed=123)
    assert sdf.name == "torus_compact"



"""Lightweight synthetic scenario library extracted from Survi.

Public API keeps compatibility with Survi's scenario loaders while avoiding
Survi-specific container types. Functions return pandas DataFrames and simple
truth callables.
"""

from .loaders import (
    list_elevation_scenarios,
    load_elevation_scenario,
    list_sdf_scenarios,
    load_sdf_scenario,
)
from .synthetic import SurfaceParams, RippleSpec, BumpSpec, make_surface_df
from .synthetic_3d import (
    Synthetic3DSurfaceFactory,
    make_asteroid_field,
    make_canal_maze,
    make_cave_network,
    make_torus,
)
from .types import ElevationSurface, SDFSurface

# Trenchfoot integration (optional dependency)
try:
    from .trenchfoot_loader import list_trenchfoot_scenarios, load_trenchfoot_scenario
    _TRENCHFOOT_AVAILABLE = True
except ImportError:
    _TRENCHFOOT_AVAILABLE = False

    def list_trenchfoot_scenarios() -> list[str]:
        """Stub when trenchfoot is not installed."""
        raise ImportError(
            "trenchfoot is not installed. "
            "Install with: pip install surface-scenarios[trenchfoot]"
        )

    def load_trenchfoot_scenario(name: str, **kwargs):
        """Stub when trenchfoot is not installed."""
        raise ImportError(
            "trenchfoot is not installed. "
            "Install with: pip install surface-scenarios[trenchfoot]"
        )

__all__ = [
    "list_elevation_scenarios",
    "load_elevation_scenario",
    "list_sdf_scenarios",
    "load_sdf_scenario",
    "SurfaceParams",
    "RippleSpec",
    "BumpSpec",
    "make_surface_df",
    "Synthetic3DSurfaceFactory",
    "make_asteroid_field",
    "make_canal_maze",
    "make_cave_network",
    "make_torus",
    "ElevationSurface",
    "SDFSurface",
    "list_trenchfoot_scenarios",
    "load_trenchfoot_scenario",
]

__version__ = "0.3.0"



import numpy as np
from survi_scenarios.synthetic_3d import Sphere, Torus, Union


def test_sphere_phi_and_gradient():
    sphere = Sphere(center=(0.0, 0.0, 0.0), radius=1.0)
    pts = np.array([[1.0, 0.0, 0.0], [0.0, 0.0, 0.0]])
    phi = sphere.phi(pts)
    assert phi[0] == 0.0  # on surface
    assert phi[1] == -1.0
    grad = sphere.gradient(pts)
    np.testing.assert_allclose(grad[0], [1.0, 0.0, 0.0])


def test_torus_phi_signs():
    torus = Torus(major_radius=0.8, minor_radius=0.25)
    pts = np.array([[1.3, 0.0, 0.0], [0.8, 0.0, 0.0]])
    phi = torus.phi(pts)
    assert phi[0] > 0  # outside
    assert phi[1] < 0  # inside


def test_union_selects_min():
    a = Sphere(center=(0, 0, 0), radius=1)
    b = Sphere(center=(2, 0, 0), radius=1)
    u = Union(a, b)
    pts = np.array([[0.0, 0.0, 0.0], [2.0, 0.0, 0.0], [1.0, 0.0, 0.0]])
    phi = u.phi(pts)
    assert phi[2] == 0.0  # mid-point lies on both surfaces
    assert phi[0] == -1.0
    assert phi[1] == -1.0
